/*
    Bantumi
    Copyright 2005, 2006 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsj
    martin@martin.st
*/

#import "cocoa.h"

#include "connection.h"
#import "BluetoothController.h"
#include "btconn.h"


void showWarning(const char *str);
NSWindow *getMainWindow();

@implementation BluetoothController
- (void) sheetDidEnd: (IOBluetoothDeviceSelectorController*) controller returnCode: (int) returnCode contextInfo: (void*) contextInfo {
	if (returnCode != kIOBluetoothUISuccess) {
		[NSApp stopModal];
		return;
	}
	device = [[controller getResults] objectAtIndex: 0];
	[device retain];

	[NSApp stopModal];
}

- (void) connect {
	IOBluetoothDeviceSelectorController *controller = [[IOBluetoothDeviceSelectorController alloc] init];
	[controller addAllowedUUID: uuid];
//	[controller setOptions: kSearchOptionsAlwaysStartInquiry];
//	[controller setOptions: kSearchOptionsDiscardCachedResults];
	[controller setPrompt: @"Connect"];
	[controller setTitle: @"Connect to game host"];
	[controller setDescriptionText: @"Select a bluetooth device hosting the game. To find devices in range of your computer, click the Search button."];
//	[[controller window] setShowsResizeIndicator: NO];
//	[controller runModal];

	[controller beginSheetModalForWindow: getMainWindow() modalDelegate: self didEndSelector: @selector(sheetDidEnd:returnCode:contextInfo:) contextInfo: NULL];

	[NSApp runModalForWindow: [controller window]];
	[controller release];

	if (device == nil)
		return;

	IOBluetoothSDPServiceRecord *record = [device getServiceRecordForUUID: uuid];
	BluetoothRFCOMMChannelID channelID;
	if ([record getRFCOMMChannelID: &channelID] != kIOReturnSuccess) {
		showWarning("Bluetooth error");
		return;
	}
	channel = nil;
	opened = NO;
	conn = new BTConnection(self, true);
	[device openRFCOMMChannelAsync: &channel withChannelID: channelID delegate: self];

	nib = [[NSNib alloc] initWithNibNamed: @"BluetoothWait" bundle: [NSBundle mainBundle]];
	[nib instantiateNibWithOwner: self topLevelObjects: nil];
	[connectingLabel setHidden: NO];
	[waitingLabel setHidden: YES];
	[progress startAnimation: self];
	[NSApp beginSheet: window modalForWindow: getMainWindow() modalDelegate: nil didEndSelector: nil contextInfo: NULL];
	cancelling = NO;
	[NSApp runModalForWindow: window];
}

- (void) accept {
	NSString *path = [[NSBundle mainBundle] pathForResource: @"SDPRecord" ofType: @"plist"];
	NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithContentsOfFile: path];
	NSMutableArray *serviceClassArray = [dict objectForKey: @"0001 - ServiceClassIDList"];
	[serviceClassArray replaceObjectAtIndex: 0 withObject: uuid];
	IOBluetoothSDPServiceRecordRef ref;
	IOBluetoothAddServiceDict((CFDictionaryRef) dict, &ref);
	IOBluetoothSDPServiceRecord *record = [IOBluetoothSDPServiceRecord withSDPServiceRecordRef: ref];
	BluetoothRFCOMMChannelID channelID = 0;
	[record getRFCOMMChannelID: &channelID];
	[record getServiceRecordHandle: &recordHandle];
	IOBluetoothObjectRelease(ref);

	conn = new BTConnection(self, false);

	channel = nil;
	opened = NO;
	notification = [IOBluetoothRFCOMMChannel registerForChannelOpenNotifications: self selector: @selector(channelOpened:channel:) withChannelID: channelID direction: kIOBluetoothUserNotificationChannelDirectionIncoming];

	nib = [[NSNib alloc] initWithNibNamed: @"BluetoothWait" bundle: [NSBundle mainBundle]];
	[nib instantiateNibWithOwner: self topLevelObjects: nil];
	[connectingLabel setHidden: YES];
	[waitingLabel setHidden: NO];
	[progress startAnimation: self];
	[NSApp beginSheet: window modalForWindow: getMainWindow() modalDelegate: nil didEndSelector: nil contextInfo: NULL];
	cancelling = NO;
	[NSApp runModalForWindow: window];
}

- (IBAction) cancel: (id) sender {
	cancelling = YES;
	[NSApp endSheet: window];
	[window orderOut: self];
	[nib release];
	nib = nil;

	[NSApp stopModal];
}

- (void) postEmptyEvent {
	// post an empty event, otherwise the run loop sticks waiting for events without terminating immediately
	NSPoint point = { 0, 0 };
	[NSApp postEvent: [NSEvent otherEventWithType: NSApplicationDefined location: point modifierFlags: 0 timestamp: 0 windowNumber: 0 context: nil subtype: 0 data1: 0 data2: 0] atStart: YES];
}

- (void) rfcommChannelOpenComplete: (IOBluetoothRFCOMMChannel*) chan status: (IOReturn) error {
	if (error != kIOReturnSuccess)
		cancelling = YES;

	if (cancelling) {
		[device release];
		device = nil;
		channel = nil;
	} else {
		channel = chan;
		[channel retain];
		opened = YES;
	}
	[self cancel: self];
	[self postEmptyEvent];

	if (error != kIOReturnSuccess) {
		char buf[100];
		sprintf(buf, "Bluetooth error %d", error);
		showWarning(buf);
	}
}

- (void) channelOpened: (IOBluetoothUserNotification*) n channel: (IOBluetoothRFCOMMChannel*) newChan {
	if (cancelling) return;

	channel = newChan;
	if (recordHandle)
		IOBluetoothRemoveServiceWithRecordHandle(recordHandle);
	recordHandle = 0;
	[notification unregister];
	notification = nil;

	[channel retain];
	[channel setDelegate: self];

	// the channel isn't open yet, wait for rfcommChannelOpenComplete
}

- (void) rfcommChannelData: (IOBluetoothRFCOMMChannel*) rfcommChannel data: (void*) dataPointer length: (size_t) dataLength {
	conn->received((const unsigned char*) dataPointer, dataLength);
}

- (void) rfcommChannelClosed: (IOBluetoothRFCOMMChannel*) rfcommChannel {
	if (!opened) {
		channel = nil;
		[self cancel: self];
		[self postEmptyEvent];
		showWarning("Connection refused");
	}
	closed = YES;
}

- (id) init {
	if ((self = [super init]) != nil) {
		channel = nil;
		device = nil;
		conn = NULL;
		released = NO;
		closed = NO;
	}
	return self;
}

- (void) shutdown {
	if (notification)
		[notification unregister];
	notification = nil;
	if (recordHandle)
		IOBluetoothRemoveServiceWithRecordHandle(recordHandle);
	recordHandle = 0;
	if (channel != nil && device == nil) {
		device = [channel getDevice];
		[device retain];
	}
	[channel setDelegate: nil];
	[channel closeChannel];
	[channel release];
	[device closeConnection];
	[device release];
	channel = nil;
	device = nil;
}

- (void) dealloc {
	[self shutdown];

	released = YES;
	if (conn && !conn->deleted)
		delete conn;

	[super dealloc];
}
@end

Connection *showBluetoothDialog(BOOL client) {
	BluetoothController *bc = [[BluetoothController alloc] init];
	bc->uuid = [IOBluetoothSDPUUID uuid32: 0x10273929];

	unsetenv("SDL_ENABLEAPPEVENTS");
	if (client) {
		[bc connect];
	} else {
		[bc accept];
	}
	setenv("SDL_ENABLEAPPEVENTS", "1", 1);

	if (bc->channel) {
		return bc->conn;
	}

	[bc shutdown];
	[bc release];
	return NULL;
}



